#!/bin/sh

. /lib/migrate-common.sh

# In RHEL8/Centos8 "multipath -u" changed to return
# DM_MULTIPATH_DEVICE_PATH=0 when device is not multipath instead of returning
# empty output - but the checks in dracut always checked for non empty path
# instead for the value "1". This was fixed upstream in dracut:
#
#   9dd9329f - modules.d: fix udev rules detection of multipath devices
#
# Apply fix here for OSs without this fix
fix_multipath() {
  for file in /etc/udev/rules.d/*.rules ; do
    sed -i -e \
      's/ENV{DM_MULTIPATH_DEVICE_PATH}=="?\*"/ENV{DM_MULTIPATH_DEVICE_PATH}=="1"/g' \
      "$file"
  done
}

fix_multipath

modprobe iscsi_ibft

if [ ! -e /sys/firmware/ibft ] ; then
  console_log "no iBFT detected - skipping iscsi boot"
  return 0
fi

console_log "iBFT detected - configuring iscsi boot"

hook_binary() {
  if [ ! -e "$1" ] ; then
    return 1
  fi
  if [ -n "$2" ] ; then
    orig="$2"
  else
    orig="$1.orig"
  fi
  mv -f "$1" "${orig}"
  cat > "$1" << EOF
#!/bin/sh
# Added for Google Cloud by Migrate for Compute Engine
. /lib/migrate-common.sh
run_hook_and_log "${orig}" "\$@"
EOF
  chmod +x "$1"
  return 0
}

# Uncomment to enable multipath debug mode
#debug_multipath=y

# shellcheck disable=SC2154
if [ -n "${debug_multipath}" ] ; then
  hook_binary "/sbin/dmsetup"
  hook_binary "/sbin/multipath" || hook_binary "/usr/sbin/multipath"
  # lvm binary name is queried in runtime and must not be renamed
  hook_binary "/sbin/lvm" "/tmp/lvm"
  hook_binary "/usr/bin/partx"

  if [ -e "/etc/multipath.conf" ] ; then
    run_and_console_log cat "/etc/multipath.conf"
  fi
  if [ -e "/etc/multipath/bindings" ] ; then
    run_and_console_log cat "/etc/multipath/bindings"
  fi
  if [ -e "/etc/multipath/wwids" ] ; then
    run_and_console_log cat "/etc/multipath/wwids"
  fi
fi

# Workaround for 56-multipath.rules (SLES 15)
# The udev rule 56-multipath uses "multipath -u" in order to test
# whether the device should be multipathed.
# However, sometimes, the first time multipath is called on
# VELOS_FE devices it returns:
#   DM_MULTIPATH_DEVICE_PATH=0
# This causes lvm to detect this device, which will prevent multipath
# from claiming the device later.
# To avoid this, all VELOS_FE devices are explicitly marked for multipath.
# This is needed even if lvm is not installed since multipath may not be
# able to claim devices after they were partitioned.
if [ -e "/usr/lib/udev/rules.d/56-multipath.rules" ] ; then
  console_log "Fixing multipath issue in udev rule: 56-multipath"
  cat > "/tmp/mpath-velos-log" << EOF
#!/bin/sh
# Added for Google Cloud by Migrate for Compute Engine
. /lib/migrate-common.sh
console_log "Marking \$1 as mpath_member"
EOF
  chmod +x "/tmp/mpath-velos-log"
  cat > "/usr/lib/udev/rules.d/57-multipath-velos.rules" << EOF
# Added for Google Cloud by Migrate for Compute Engine
ACTION=="add", \
SUBSYSTEM=="block", \
ENV{DEVTYPE}!="partition", \
ATTRS{vendor}=="VELOS_FE", \
ATTRS{model}=="VDISK           ", \
ENV{DM_MULTIPATH_DEVICE_PATH}="1", \
ENV{ID_FS_TYPE}="mpath_member", \
RUN+="/usr/bin/partx -d --nr 1-1024 \$kernel", \
RUN+="/tmp/mpath-velos-log \$kernel"
EOF
  run_and_console_log udevadm control --reload-rules
fi

# Workaround for 62-multipath.rules (Centos 7 / Centos 8)
if [ -e "/usr/lib/udev/rules.d/62-multipath.rules" ] ; then
  console_log "Fixing multipath issue in udev rule: 62-multipath"
  cat > "/tmp/mpath-velos-log" << EOF
#!/bin/sh
# Added for Google Cloud by Migrate for Compute Engine
. /lib/migrate-common.sh
console_log "Marking \$1 as mpath_member"
EOF
  chmod +x "/tmp/mpath-velos-log"
  if grep -q "multipath -c" "/usr/lib/udev/rules.d/62-multipath.rules" ; then
    # Centos 7 uses multipath -c - old rules file
    cat > "/usr/lib/udev/rules.d/63-multipath-velos.rules" << EOF
# Added for Google Cloud by Migrate for Compute Engine
ACTION=="add", \
SUBSYSTEM=="block", \
ENV{DEVTYPE}!="partition", \
ATTRS{vendor}=="VELOS_FE", \
ATTRS{model}=="VDISK           ", \
ENV{DM_MULTIPATH_DEVICE_PATH}="1", \
ENV{DM_MULTIPATH_WIPE_PARTS}="1", \
ENV{ID_FS_TYPE}="mpath_member", \
RUN+="/sbin/partx -d --nr 1-1024 \$env{DEVNAME}", \
RUN+="/tmp/mpath-velos-log \$env{DEVNAME}"
EOF
  else
    # Centos 8 uses "multipath -u" - new rules file
    cat > "/usr/lib/udev/rules.d/63-multipath-velos.rules" << EOF
# Added for Google Cloud by Migrate for Compute Engine
ACTION=="add", \
SUBSYSTEM=="block", \
ENV{DEVTYPE}!="partition", \
ATTRS{vendor}=="VELOS_FE", \
ATTRS{model}=="VDISK           ", \
ENV{DM_MULTIPATH_DEVICE_PATH}="1", \
ENV{ID_FS_TYPE}="mpath_member", \
RUN+="/usr/sbin/partx -d --nr 1-1024 \$env{DEVNAME}", \
RUN+="/tmp/mpath-velos-log \$env{DEVNAME}"
EOF
  fi
  run_and_console_log udevadm control --reload-rules
fi

{
  # Forces invocation of "ibft_to_cmdline()" by dracut
  # * Rename iface to ibftX
  # * Adds ip=ibft:dhcp to effective cmdline
  # Forces addition of rd.neednet to effective cmdline (existence of ip=
  # lines)
  echo "rd.iscsi.ibft ip=ibft"
  # Forces parse-iscsiroot to invoke iscsiroot (iscsi boot)
  echo "rd.iscsi.firmware iscsi_firmware"
  # Forces multipath
  echo "rd.multipath"
  # Forces retry to secondary iscsi target from ibft if primary is offline -
  # we need more than one retry for that
  echo "iscsi_param=node.session.initial_login_retry_max=4"
  echo "rd.iscsi.param=node.session.initial_login_retry_max=4"
} >> /etc/cmdline
# reread cmdline
unset CMDLINE

# Directly use udev to trigger iscsi-ibft-routes
# By choosing a name that starts with "00-" we ensure that iscsi-ibft-routes
# will run on every interface (except 'lo') before ifup does.
# iscsi-ibft-routes will set the the interface settings (using iscsistart -N)
# or at least set the routes correctly before ifup has a chance to set the rest
# of the settings using dhcp.
# After that, the iscsi session will start without failing.
# This works on all versions of dracut (and opensuse-dracut)
{
  runcmd="/sbin/initqueue --onetime --unique "
  runcmd="${runcmd} --name 00-iscsi-ibft-routes-\$env{INTERFACE} "
  runcmd="${runcmd} /sbin/iscsi-ibft-routes \$env{INTERFACE}"
  rule="SUBSYSTEM==\"net\", "
  rule="${rule} ACTION==\"add\", "
  rule="${rule} ENV{INTERFACE}!=\"lo\", "
  rule="${rule} RUN+=\"${runcmd}\""
  echo "${rule}"
} >> /etc/udev/rules.d/00-migrate-for-gce-hook.rules

# In RHEL7/Centos7 dracut creates the following file in the pre-pivot stage on
# the real root (/sysroot):
#
#   /etc/sysconfig/network-scripts/ifcfg-ibft0
#
# This file contains HWADDR={mac address of ibft0 interface}
# Later on, when the system boots, udev gets loaded and starts processing its
# rules.
# One of them is /usr/lib/udev/rules.d/60-net.rules which invokes the
# /lib/udev/rename_device helper utility. This helper utility
# iterates on all /etc/sysconfig/network-scripts/ifcfg-*suffix* files - if a
# file contains HWADDR with a mac address of an existing interface - it renames
# this interface to the given name in the ifcfg file.
# After detach the interface becomes eth0 and later renamed to ibft0 since the
# mac address of the interface does not change.
#
# Avoid this behavior by removing the write-ifcfg hook in dracut which generates
# the /etc/sysconfig/network-scripts/ifcfg-ibft0 file in the first place.
for hook in /lib/dracut/hooks/pre-pivot/*-write-ifcfg.sh ; do
  if [ ! -f "${hook}" ] ; then
    continue
  fi
  console_log "Deleting ${hook}"
  rm "${hook}"
done

fix_iscsiroot() {
  # Some versions of SLES12 (SP3, SP4) have a bug in their iscsrioot script
  # It seems that the patch from the following commit was applied wrongfully:
  #   https://github.com/dracutdevs/dracut/commit/0d5597044fb8690997d
  # The first part of the commit which replaces "case" with "if" is applied
  # but the second part of the commit which replaces "esac" with "fi" is not
  # applied.
  # This results in iscsiroot script failing with:
  #   dracut-initqueue[551]: /sbin/iscsiroot: line 254: syntax error near unexpected token `;;\'\r\n
  # We then fail to iscsi boot.
  # Since this bug seems to be affecting many images we fix it here by applying
  # the rest of the patch - replacing "esac" with "fi".

  # shellcheck disable=SC2016
  signature1='if \[ "$target" = "$iscsi_target_name" \]; then'
  signature2='(?s)        ;;\n        \*\)\n        ;;\n        esac\n'
  iscsiroot_file=/sbin/iscsiroot
  fixed_file=/tmp/fixed_iscsiroot

  if [ ! -x ${iscsiroot_file} ] ; then
    return 0
  fi

  if ! grep -q    "${signature1}" ${iscsiroot_file} ||
     ! grep -qPzo "${signature2}" ${iscsiroot_file} ; then
     return 0
  fi

  console_log "Detected bad version of iscsiroot - trying to fix..."
  pattern='s/        ;;\n        \*)\n        ;;\n        esac/        fi/'
  # shellcheck disable=SC2016
  escape_lines_pattern='1h;2,$H;$!d;g'
  sed -e "${escape_lines_pattern}" \
      -e "${pattern}" ${iscsiroot_file} > ${fixed_file}

  if ! grep -q "${signature1}" ${fixed_file} ||
       grep -qPzo "${signature2}" ${fixed_file} ; then
    console_log "Failed fixing iscsiroot"
    return 1
  fi

  if ! chmod +x ${fixed_file} ; then
    console_log "Failed setting ${fixed_file} as executable"
    return 1
  fi

  if ! mv -f ${fixed_file} ${iscsiroot_file} ; then
    console_log "Failed replacing ${iscsiroot_file} with ${fixed_file}"
    return 1
  fi

  console_log "Successfully fixed iscsiroot"

  return 0
}

fix_iscsiroot
