#!/bin/bash

. /opt/migrate-for-gce/functions.sh

is_ipv4() {
  ip=${1:-1.2.3.4}
  if expr "$ip" : \
      '[0-9][0-9]*\.[0-9][0-9]*\.[0-9][0-9]*\.[0-9][0-9]*$' \
      >/dev/null; then
    OIFS=$IFS
    IFS=.
    # shellcheck disable=SC2086
    set $ip
    for quad in 1 2 3 4; do
      if eval [ \$$quad -gt 255 ]; then
        IFS=$OIFS
        return 1
      fi
    done
    IFS=$OIFS
    return 0
  else
    return 1
  fi
}

is_loopback() {
  ip=$1
  ip_first_section="${ip%%.*}"
  if [ "${ip_first_section}" = "127" ]; then
    return 0
  fi
  return 1
}

get_ip() {
  get_instance_metadata "network-interfaces/0/ip"
}

get_short_hostname() {
  fqdn=$(hostname)
  echo "${fqdn%%.*}"
}

get_persistent_hostname() {
  if [ -f /etc/HOSTNAME ]; then
    cut -d "." -f 1 /etc/HOSTNAME
  elif [ -f /etc/hostname ]; then
    cut -d "." -f 1 /etc/hostname
  fi
}

should_replace_ip() {
  ip=$1
  suffix=$2
  actual_ip=$3
  hostname=$4
  hostname_from_file=$5

  if ! is_ipv4 "${ip}" ; then
    return 1
  fi
  if is_loopback "${ip}" ; then
    return 1
  fi
  if [ "${ip}" = "${actual_ip}" ] ; then
    return 1
  fi
  if echo "${suffix}" | grep -q "${hostname}" ; then
    return 0
  fi
  if [ -n "${hostname_from_file}" ] ; then
    if echo "${suffix}" | grep -q "${hostname_from_file}" ; then
      return 0
    fi
  fi
  if echo "${suffix}" | grep -q "localhost" ; then
    return 0
  fi
  return 1
}

hosts_file=/etc/hosts
hosts_backup_file=/etc/hosts.pre-gce

if [ "$1" = "revert" ] ; then
  if [ ! -f "${hosts_backup_file}" ] ; then
    exit 0
  fi
  mv -f "${hosts_backup_file}" "${hosts_file}"
  exit $?
fi

# First stage - before networking
if [ "$1" = "metadata" ] ; then
  # Backup the hosts file
  if [ ! -f "${hosts_backup_file}" ] ; then
    cp "${hosts_file}" "${hosts_backup_file}"
  fi
  # Add the metadata server
  if ! grep -q metadata.google.internal "${hosts_file}" ; then
    {
      echo ""
      echo "169.254.169.254  metadata.google.internal"
    } >> "${hosts_file}"
  fi
  exit 0
fi

if [ "$1" != "ip" ] ; then
  # Bad option
  exit 1
fi

# Second stage - after networking
hostname=$(get_short_hostname)
if [ -z "${hostname}" ]; then
  migrate_log "error - unable to get the host name"
  exit 1
fi
hostname_from_file=$(get_persistent_hostname)
migrate_log "Using hostnames = ${hostname}, ${hostname_from_file}"

actual_ip=$(get_ip)
if [ -z "${actual_ip}" ] ; then
  migrate_log "error - unable to detect local ip"
  exit 1
fi
migrate_log "Detected local IP = ${actual_ip}"

modified=0
temp_file=$(mktemp)

if [ -n "${hostname}" ] && [ -n "${actual_ip}" ] ; then

  while read -r line; do
    ip=$(echo "${line}" | awk '{print $1}')
    suffix=$(echo "${line}" | awk '{for (i=2; i<=NF; i++) print $i}')

    if should_replace_ip "${ip}" "${suffix}" "${actual_ip}" \
        "${hostname}" "${hostname_from_file}"; then
      migrate_log "Found obsolete IP for ${suffix}, update: ${ip} => ${actual_ip}"
      echo "${actual_ip}" "${suffix}" >> "${temp_file}"
      modified=1
    else
      echo "${line}" >> "${temp_file}"
    fi

  done < ${hosts_file}

else

  cat ${hosts_file} > "${temp_file}"

fi

if [ $modified -eq 1 ] ; then
  mv -f "${temp_file}" "${hosts_file}"
fi

rm -f "${temp_file}"

exit 0
