#!/bin/bash


read -d '' JSON_HEADER <<- EOF
{
  "header": {
	"version": "1",
	"type": "json"
  }
}
EOF

read -d '' JSON_RPC_TEMPLATE <<- EOF
{
  "jsonrpc": "2.0",
  "method": "heart-beat",
  "params": {
	"type": "%s",
	"uptime": "%s"
  },
  "id": %d
}
EOF

function xxd_missing
{
	command -v xxd >/dev/null 2>&1
	echo "$?"
}

FIXED_RPC_HEADER_SIZE=512
BUFFER_ID=1
XXD_MISSING=$(xxd_missing)
TIMED_OUT=137

ISCSIADM="iscsiadm"
PIDFILE=/var/run/migrate-for-gce-keep-alive.pid

export PATH=${PATH}:/sbin:/bin:/usr/sbin:/usr/bin


function async_log
{
	( timeout -s 9 10 logger [`date +"%F %T"`] MigrateForGCE: $* ) &
}

function console_log
{
	echo [`date +"%F %T"`] MigrateForGCE: $* > /dev/kmsg
}

function error_log
{
	echo [`date +"%F %T"`] ERR: $* 1>&2
}

function createJsonRequest
{
	keepalive_type=$1
	requestId=$2
	headerLength=${#JSON_HEADER}
	let "paddingSize = $FIXED_RPC_HEADER_SIZE - $headerLength"
	uptimeSecondsAgo="`cat /proc/uptime | cut -f1 -d.` seconds ago"
	uptimeIso8601=`date +%Y-%m-%dT%H:%M:%S%z -d "$uptimeSecondsAgo"`
	rpcJson=`printf "$JSON_RPC_TEMPLATE" "$keepalive_type" "$uptimeIso8601" "$requestId"`

	#error_log "Creating rpc request paddingSize=$paddingSize, uptimeSecondsAgo=$uptimeSecondsAgo uptimeIso8601=$uptimeIso8601 headerLength=$headerLength"

	jsonHeaderHex=`printf "$JSON_HEADER" | hexdump -ve '/1 "%02x"'`
	rpcJsonHex=`echo -n "$rpcJson" | hexdump -ve '/1 "%02x"'`
	paddingHex=`printf '00%.0s' $(seq 1 $paddingSize)`

	echo -n "${jsonHeaderHex}${paddingHex}${rpcJsonHex}"
}

function sendKeepAliveRpc
{
	local keepalive_type velos_dev running_id

	keepalive_type=$1
	velos_dev=$2
	running_id=$3

	hexRequest=`createJsonRequest "${keepalive_type}" "${running_id}"`
	if [ $XXD_MISSING -eq 1 ]; then
		echo -n "$hexRequest" | sed 's/\([0-9A-F]\{2\}\)/\\\\\\x\1/gI' | xargs printf | sg_write_buffer --id=${BUFFER_ID} -r --mode=1 "$velos_dev"
	else
		#prefer xxd if possible
		echo -n "$hexRequest" | xxd -r -p | sg_write_buffer --id=${BUFFER_ID} -r --mode=1 "$velos_dev"
	fi
}

function safe_mp_status()
{
	timeout -s 9 30 /sbin/multipath -ll $*
}

function mpath_debug_output
{
	local dev scsi_dev paths

	if [ `grep -c "Ubuntu 14" /etc/issue` -eq 1 ]; then
		# Ubuntu 14 fails to run cmds like 'multipath -ll /dev/mapper/mpath0'.
		scsi_dev=$(safe_mp_status | grep -A3 dm-0 | tail -1 | awk '{print $4}')
		if [ ! -z "${scsi_dev}" ] ; then
			paths=$(safe_mp_status /dev/${scsi_dev} | grep '`- ')
		fi
	else
		dev=$(safe_mp_status | grep dm-0 | awk '{print $1}')
		if [ ! -z "${dev}" ] ; then
			paths=$(safe_mp_status /dev/mapper/${dev} | grep '`- ')
		fi
	fi

	console_log $dev $paths
	async_log $dev $paths
}

function find_velos_lun0_devices
{
	local velos_blk_devs

	velos_blk_devs=`ls -l /dev/disk/by-path | egrep "velostrata.iscsi.base" | egrep lun-0 | grep -v "\-part" | cut -d/ -f3`
	echo ${velos_blk_devs}
}

function send_tur
{
	local vendor_bits velos_dev keepAliveType running_id

	vendor_bits=$1
	keepAliveType=$2
	running_id=$3
	velos_devs=$(find_velos_lun0_devices)

	if [ -z "${velos_devs}" ] ; then
		async_log "could not find Migrate for GCE edge VM device"
	else
		for velos_dev in ${velos_devs}; do
			velos_dev_full_path=/dev/${velos_dev}
			rpcOutput=`sendKeepAliveRpc "${keepAliveType}" "${velos_dev_full_path}" "${running_id}" 2>&1`
			rpcRc=$?
			sg_raw -r 1k ${velos_dev_full_path} 00 00 00 00 00 ${vendor_bits} > /dev/null 2>&1
			turRc=$?
			if [ ${turRc} -ne 0 ] || [ ${rpcRc} -ne 0 ]; then
				active=`multipath -ll | grep -B1 ${velos_dev} | grep -c "status=active"`

				# If we failed to send a TUR to an inactive path, don't print an error log
				if [ ${active} -ne 0 ]; then
					async_log "could not send keep-alive message to Migrate for GCE edge VM via ${velos_dev} (${turRc}/${rpcRc})"
				fi
			fi
		done
	fi
}

function send_first_tur
{
	running_id=$1
	# Send a TUR cmd. Set the vendor-specific bits to '10'.
	send_tur "80" "post-boot" ${running_id}
}

function send_regular_tur
{
	running_id=$1
	# Send a TUR cmd. Set the vendor-specific bits to '11'.
	send_tur "C0" "keep-alive" ${running_id}
}

function cleanup_iscsi_tgt
{
	ip=$1
	tgt=$2

	async_log "cleanup ${ip}::::${tgt}"
	${ISCSIADM} --mode node --portal ${ip} --target ${tgt} --logout
	${ISCSIADM} --mode node --portal ${ip} --target ${tgt} -o delete
}

function reconnect_to_iscsi_tgt
{
	ip=$1
	tgt=$2

	async_log "reconnecting to ${ip}::::${tgt}"
	if timeout 30 ${ISCSIADM} --mode discovery --portal ${ip} --type sendtargets ; then
		${ISCSIADM} --mode node --portal ${ip} -o update -n node.startup -v manual
		${ISCSIADM} --mode node --portal ${ip} --target ${tgt} --login
	fi
}

function refresh_paths()
{
	local tgt_base ip iqn mp_result

	safe_mp_status &>/dev/null
	mp_result=$?

	# Connect to targets that were missing during boot
	for tt in 0 1 ; do
		tgt_base=/sys/firmware/ibft/target${tt}
		if [ -e ${tgt_base} ] ; then
			ip=`cat ${tgt_base}/ip-addr`
			iqn=`cat ${tgt_base}/target-name`

			# Sometimes the path is down, but the device is still there.
			# In this case rediscovering the portal helps
			if [ ${mp_result} -eq ${TIMED_OUT} ]; then
				timeout 30 ${ISCSIADM} --mode discovery --portal ${ip} --type sendtargets
			fi

			if  ! ls /dev/disk/by-path/*${ip}* > /dev/null 2>&1 ; then
				cleanup_iscsi_tgt ${ip} ${iqn}
				reconnect_to_iscsi_tgt ${ip} ${iqn}
			fi
		fi
	done

	mpath_debug_output
}

function disable_ssi_luns()
{
	${ISCSIADM} --mode node --target iqn.2014-05.com.velostrata.iscsi.base:WindowsServicingData -o update -n node.startup -v manual
}

function check_multipath()
{
	if [ `grep -c "Ubuntu 14" /etc/issue` -eq 0 ]; then
		return
	fi

	/sbin/multipathd show paths | grep "^[0-9]*:[0-9]*:[0-9]*:[0-9]*" | while read -r line ; do
		found=`echo ${line} | grep -c orphan`
		if [ ${found} -eq 1 ]; then
			logger "MigrateForGCE: found an orphan - ${line}"

			logger "MigrateForGCE: restarting multipath service"
			service multipath-tools restart
			return
		fi
	done
}


echo $$ > $PIDFILE

if [ ! -e /sys/firmware/ibft ]; then
	logger "MigrateForGCE: VM is not in cloud, skipping keep-alive messages"
	sleep infinity
	rm -f $PIDFILE
	exit 0
fi

disable_ssi_luns
is_first_tur=1
running_rpc_id=0

while [ 1 ]; do
	refresh_paths &
	refresh_paths_pid=$!

	# workaround for HA-4612
	check_multipath

	running_rpc_id=$((running_rpc_id+1))

	if [ ${is_first_tur} -eq 1 ]; then
		async_log "starting keep-alive messages to Migrate for GCE edge VM"
		send_first_tur $running_rpc_id &
		send_tur_pid=$!
		is_first_tur=0
	else
		send_regular_tur $running_rpc_id &
		send_tur_pid=$!
	fi

	sleep 60
	for async_pid in ${send_tur_pid} ${refresh_paths_pid}; do
	    kill -0 ${async_pid} &>/dev/null
	    process_is_running=$?
	    if [ ${process_is_running} -eq 0 ] ; then
	        async_log "keep alive subprocess is stuck. Terminating"
	        kill -9 ${async_pid}
	    fi
	done
done

rm -f $PIDFILE
exit 1
